@@ -28,6 +28,13 @@ from typing import (
 
 import polars as pl
 
+try:
+    import zoneinfo
+except ImportError:
+    # Zoneinfo was introduced in python 3.9
+    from backports import zoneinfo
+from dateutil import parser
+
 from chalk.features.feature_field import Feature
 from chalk.features.feature_wrapper import FeatureWrapper, unwrap_feature
 from chalk.features.filter import Filter, TimeDelta
@@ -57,6 +64,16 @@ def _polars_data_frames_equal(a: pl.DataFrame, b: pl.DataFrame):
     return a.frame_equal(b)
 
 
+def _iso_parse(x: str, expected_tz: datetime.tzinfo | None) -> datetime.datetime:
+    """This function converts the input into an expected timezone, and drops it, so we can cast it in polars with with_time_zone"""
+    parsed = parser.isoparse(x)
+    if parsed.tzinfo is not None and expected_tz is not None:
+        # Convert the timezone
+        parsed = parsed.astimezone(expected_tz)
+    # Now drop the timezone
+    return parsed.replace(tzinfo=None)
+
+
 class DataFrameMeta(type):
     def __getitem__(cls, item: Any) -> Type[DataFrame]:
         from chalk.features.feature_set import Features
@@ -292,13 +309,23 @@ class DataFrame(metaclass=DataFrameMeta):
                                         )
                                     else:
                                         try:
-                                            if element_as_series.dtype == pl.Utf8 and polars_dtype == pl.Datetime:
-                                                element_as_series = element_as_series.str.strptime(
-                                                    pl.Datetime, fmt="+", strict=False
-                                                ).cast(polars_dtype)
+                                            if (
+                                                isinstance(element_as_series.dtype, type)
+                                                and issubclass(element_as_series.dtype, pl.Utf8)
+                                                and isinstance(polars_dtype, pl.Datetime)
+                                            ):
+                                                if polars_dtype.tz is None:
+                                                    element_as_series = element_as_series.apply(
+                                                        lambda x: None if x is None else _iso_parse(x, None),
+                                                    ).dt.with_time_zone(None)
+                                                else:
+                                                    tzinfo = zoneinfo.ZoneInfo(polars_dtype.tz)
+                                                    element_as_series = element_as_series.apply(
+                                                        lambda x: None if x is None else _iso_parse(x, tzinfo),
+                                                    ).dt.with_time_zone(polars_dtype.tz)
                                             else:
                                                 element_as_series = element_as_series.cast(polars_dtype)
-                                        except pl.ComputeError as e:
+                                        except (pl.ComputeError, pl.InvalidOperationError) as e:
                                             # If operating on a lazy frame, we won't get the exception until we call .compute()
                                             raise TypeError(
                                                 f"Feature `{str(k)}` at index {i} with value '{element}' could not be converted to dtype `{polars_dtype.string_repr()}`"
@@ -338,14 +365,26 @@ class DataFrame(metaclass=DataFrameMeta):
                         )
                     else:
                         try:
-                            if expected_dtype == pl.Datetime and actual_dtype == pl.Utf8:
-                                underlying = underlying.with_columns(
-                                    [
+                            if isinstance(expected_dtype, pl.Datetime) and actual_dtype == pl.Utf8:
+                                tzinfo = None if expected_dtype.tz is None else zoneinfo.ZoneInfo(expected_dtype.tz)
+                                if expected_dtype.tz is None:
+                                    col = (
                                         pl.col(root_fqn)
-                                        .str.strptime(pl.Datetime, fmt="+", strict=False)
-                                        .cast(expected_dtype)
-                                    ]
-                                )
+                                        .apply(
+                                            lambda x: None if x is None else _iso_parse(cast(str, x), None),
+                                        )
+                                        .dt.with_time_zone(None)
+                                    )
+                                else:
+                                    tzinfo = zoneinfo.ZoneInfo(expected_dtype.tz)
+                                    col = (
+                                        pl.col(root_fqn)
+                                        .apply(
+                                            lambda x: None if x is None else _iso_parse(cast(str, x), tzinfo),
+                                        )
+                                        .dt.with_time_zone(expected_dtype.tz)
+                                    )
+                                underlying = underlying.with_columns([col])
                             else:
                                 underlying = underlying.with_columns([pl.col(root_fqn).cast(expected_dtype)])
                         except pl.ComputeError as e:
