@@ -1,4 +1,5 @@
 import ast
+import copy
 import functools
 import inspect
 import re
@@ -28,6 +29,8 @@ from chalk.features.feature_set import Features, FeatureSetBase
 from chalk.features.feature_wrapper import FeatureWrapper
 from chalk.features.tag import Tags
 from chalk.serialization.parsed_annotation import ParsedAnnotation
+from chalk.streams import Windowed
+from chalk.streams._windows import WindowedMeta
 from chalk.utils.collections import ensure_tuple
 from chalk.utils.duration import Duration, timedelta_to_duration
 from chalk.utils.metaprogramming import MISSING, create_fn, field_assign, set_new_attribute
@@ -239,8 +242,14 @@ def _get_field(
     default = getattr(cls, annotation_name, MISSING)
 
     if isinstance(default, Feature):
+        # The feature was set like x: int = Feature(...)
         f = default
+    elif isinstance(default, Windowed):
+        # The feature was set like x: Windowed[int] = windowed()
+        # Convert it to a Feature
+        f = default.to_feature(bucket=None)
     else:
+        # The feature was net set explicitly
         if isinstance(default, types.MemberDescriptorType):
             # This is a field in __slots__, so it has no default value.
             default = MISSING
@@ -436,8 +445,38 @@ def _process_class(
     max_staleness: Optional[Duration],
     namespace: str,
 ) -> Type[T]:
-    cls_annotations = cls.__dict__.get("__annotations__", {})
+    raw_cls_annotations = cls.__dict__.get("__annotations__", {})
 
+    cls_annotations = {}
+    for name, annotation in raw_cls_annotations.items():
+        if isinstance(annotation, Windowed):
+            # NOTE: For Windowed resolvers, both the Annotation and the value are instances of Windowed, unlike normal features
+            # whose annotation is the underlying type, and the value is an instance of FeatureWrapper
+            # So both `annotation` and `wind` should be instances of Windowed
+            # In the future, we should use a subclass of Windowed, rather than an instance, for the type annotation, similar to
+            # what we do for Features
+            wind = getattr(cls, name, None)
+            if wind is None or not isinstance(wind, Windowed):
+                raise TypeError(
+                    (
+                        f"Windowed feature '{namespace}.{name}' is missing windows. "
+                        f"To create a windowed feature, use "
+                        f"'{name}: Windowed[{annotation._kind.__name__}] = windowed('10m', '30m')."
+                    )
+                )
+            wind._name = name
+            annotation._buckets = wind._buckets
+            for bucket in wind._buckets:
+                # Make pseudofeatures for each bucket of the window
+                feat = wind.to_feature(bucket=bucket)
+                setattr(cls, feat.name, feat)
+                # For the psuedofeatures, which track an individual bucket, the correct annotation is the
+                # underlying annotation, not Windowed[underlying], since it's only one value
+                cls_annotations[feat.name] = wind._kind
+
+        cls_annotations[name] = annotation
+
+    cls.__annotations__ = cls_annotations
     if cls.__module__ in sys.modules:
         _globals = sys.modules[cls.__module__].__dict__
     else:
@@ -445,12 +484,14 @@ def _process_class(
 
     # Find feature times that weren't annotated.
     for (name, member) in inspect.getmembers(cls):
+        if name not in cls_annotations and isinstance(member, Windowed):
+            raise TypeError(f"Windowed feature '{namespace}.{name}' is missing an annotation, like 'Windowed[str]'")
         if name not in cls_annotations and isinstance(member, Feature):
             # All feature types need annotations, except for datetimes, which we can automatically infer
             if member.typ is not None and member.typ.is_parsed and issubclass(member.typ.underlying, datetime):
                 cls_annotations[name] = datetime
             else:
-                raise TypeError(f"Member {name} is missing an annotation")
+                raise TypeError(f"Feature '{namespace}.{name}' is missing an annotation")
 
     # If we pass this function something that isn't a class, it could raise
     try:
@@ -471,6 +512,7 @@ def _process_class(
         )
         for name in cls_annotations
     ]
+
     # If there is no timestamp feature, synthesize one
     ts_feature: Optional[Feature] = next((x for x in cls_fields if x.is_feature_time), None)
     if ts_feature is None:
@@ -497,6 +539,7 @@ def _process_class(
             class_max_staleness=max_staleness,
         )
         cls_fields.append(ts_feature)
+
     set_new_attribute(
         cls=cls,
         name="__init__",
