@@ -1,8 +1,7 @@
-import ast
 import dataclasses
 import inspect
-import textwrap
 from dataclasses import dataclass
+from inspect import Parameter
 from typing import (
     Any,
     Callable,
@@ -16,6 +15,7 @@ from typing import (
     Type,
     TypeVar,
     Union,
+    cast,
     get_args,
     get_origin,
     get_type_hints,
@@ -25,21 +25,31 @@ from typing import (
 from pydantic import BaseModel
 from typing_extensions import ParamSpec
 
-from chalk.df.ast_parser import convert_slice, eval_converted_expr
+from chalk.features import FeatureSetBase
 from chalk.features.dataframe import DataFrame
 from chalk.features.feature_field import Feature
 from chalk.features.feature_set import Features
 from chalk.features.feature_wrapper import FeatureWrapper, unwrap_feature
 from chalk.features.filter import Filter
 from chalk.features.tag import Environments, Tags
+from chalk.serialization.parsed_annotation import ParsedAnnotation
 from chalk.sql.base.protocols import BaseSQLSourceProtocol
 from chalk.state import StateWrapper
+from chalk.streams import StreamSource, Windowed, get_name_with_duration
+from chalk.streams.types import (
+    StreamResolverParam,
+    StreamResolverParamKeyedState,
+    StreamResolverParamMessage,
+    StreamResolverParamMessageWindow,
+    StreamResolverSignature,
+)
+from chalk.utils import MachineType
+from chalk.utils.annotation_parsing import ResolverAnnotationParser
 from chalk.utils.collection_type import GenericAlias
 from chalk.utils.collections import ensure_tuple
 from chalk.utils.duration import Duration, ScheduleOptions
 from chalk.utils.environment_parsing import env_var_bool
 
-MachineType = str
 T = TypeVar("T")
 P = ParamSpec("P")
 
@@ -75,23 +85,23 @@ class Cron:
     trigger_downstream: Optional[bool] = True
 
 
-def _get_expected_features_list(output: Type[Features]) -> Sequence[Feature]:
-    if issubclass(output, Features):
-        output = output.features
-
-    if len(output) == 1 and issubclass(output[0], DataFrame):
-        return output[0].columns
+def _flatten_features(output: Optional[Type[Features]]) -> Sequence[Feature]:
+    if output is None:
+        return []
+    features = output.features
+    if len(features) == 1 and isinstance(features[0], type) and issubclass(features[0], DataFrame):
+        return features[0].columns
     else:
-        return output
+        return features
 
 
-class Resolver(Protocol):
+class Resolver(Protocol[P, T]):
     function_definition: str
     fqn: str
     filename: str
     inputs: List[Feature]
     output: Type[Features]
-    fn: Callable
+    fn: Callable[P, T]
     environment: Optional[List[str]]
     tags: Optional[List[str]]
     max_staleness: Optional[Duration]
@@ -105,27 +115,31 @@ class Resolver(Protocol):
     def __hash__(self):
         return hash(self.fqn)
 
+    def __call__(self, *args: P.args, **kwds: P.kwargs) -> T:
+        return self.fn(*args, **kwds)
 
-def _process_call(result, *, declared_output: Any):
+    @property
+    def __name__(self):
+        return self.fn.__name__
+
+
+def _process_call(result: T, *, declared_output: Optional[Type[Features]]) -> T:
     from chalk.sql import StringChalkQuery
     from chalk.sql.integrations.chalk_query import ChalkQuery
 
     if isinstance(result, (ChalkQuery, StringChalkQuery)):
-        return result.execute_internal(expected_features=_get_expected_features_list(declared_output))
+        return result.execute_internal(expected_features=_flatten_features(declared_output))
 
     return result
 
 
-class SinkResolver:
+class SinkResolver(Resolver[P, T]):
     registry: "List[SinkResolver]" = []
 
     def __eq__(self, other):
         return isinstance(other, SinkResolver) and self.fqn == other.fqn
 
-    def __hash__(self):
-        return hash(self.fqn)
-
-    def __call__(self, *args, **kwargs):
+    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
         return _process_call(self.fn(*args, **kwargs), declared_output=None)
 
     def __init__(
@@ -136,7 +150,7 @@ class SinkResolver:
         doc: Optional[str],
         inputs: List[Feature],
         output: Type[Features],
-        fn: Callable,
+        fn: Callable[P, T],
         environment: Optional[List[str]],
         tags: Optional[List[str]],
         machine_type: Optional[MachineType],
@@ -152,6 +166,9 @@ class SinkResolver:
         self.inputs = inputs
         self.output = output
         self.fn = fn
+        self.__module__ = fn.__module__
+        self.__doc__ = fn.__doc__
+        self.__annotations__ = fn.__annotations__
         self.environment = environment
         self.tags = tags
         self.doc = doc
@@ -161,31 +178,23 @@ class SinkResolver:
         self.max_delay = max_delay
         self.upsert = upsert
         self.integration = integration
+        self.default_args = []
+        self.max_staleness = None
+        self.state = None
 
     def __repr__(self):
         return f"SinkResolver(name={self.fqn})"
 
 
-class OnlineResolver(Resolver):
+class OnlineResolver(Resolver[P, T]):
     cron: Union[ScheduleOptions, Cron]
 
     def __eq__(self, other):
         return isinstance(other, OnlineResolver) and self.fqn == other.fqn
 
-    def __hash__(self):
-        return hash(self.fqn)
-
-    def __call__(self, *args, **kwargs):
+    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
         return _process_call(self.fn(*args, **kwargs), declared_output=self.output)
 
-    @property
-    def __name__(self):
-        return self.fn.__name__
-
-    @property
-    def __module__(self):
-        return self.fn.__module__
-
     def __init__(
         self,
         function_definition: str,
@@ -194,7 +203,7 @@ class OnlineResolver(Resolver):
         doc: Optional[str],
         inputs: List[Feature],
         output: Type[Features],
-        fn: Callable,
+        fn: Callable[P, T],
         environment: Optional[List[str]],
         tags: Optional[List[str]],
         max_staleness: Optional[Duration],
@@ -210,6 +219,9 @@ class OnlineResolver(Resolver):
         self.inputs = inputs
         self.output = output
         self.fn = fn
+        self.__module__ = fn.__module__
+        self.__doc__ = fn.__doc__
+        self.__annotations__ = fn.__annotations__
         self.environment = environment
         self.tags = tags
         self.max_staleness = max_staleness
@@ -224,24 +236,13 @@ class OnlineResolver(Resolver):
         return f"OnlineResolver(name={self.fqn})"
 
 
-class OfflineResolver(Resolver):
+class OfflineResolver(Resolver[P, T]):
     def __eq__(self, other):
         return isinstance(other, OfflineResolver) and self.fqn == other.fqn
 
-    def __hash__(self):
-        return hash(self.fqn)
-
-    def __call__(self, *args, **kwargs):
+    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
         return _process_call(self.fn(*args, **kwargs), declared_output=self.output)
 
-    @property
-    def __name__(self):
-        return self.fn.__name__
-
-    @property
-    def __module__(self):
-        return self.fn.__module__
-
     def __init__(
         self,
         function_definition: str,
@@ -250,7 +251,7 @@ class OfflineResolver(Resolver):
         doc: Optional[str],
         inputs: List[Feature],
         output: Type[Features],
-        fn: Callable,
+        fn: Callable[P, T],
         environment: Optional[List[str]],
         tags: Optional[List[str]],
         max_staleness: Optional[Duration],
@@ -268,6 +269,9 @@ class OfflineResolver(Resolver):
         self.inputs = inputs
         self.output = output
         self.fn = fn
+        self.__module__ = fn.__module__
+        self.__doc__ = fn.__doc__
+        self.__annotations__ = fn.__annotations__
         self.environment = environment
         self.tags = tags
         self.max_staleness = max_staleness
@@ -281,56 +285,17 @@ class OfflineResolver(Resolver):
 
 
 @dataclasses.dataclass(frozen=True)
-class ResolverParseResult:
+class ResolverParseResult(Generic[P, T]):
     fqn: str
     inputs: List[Feature]
     state: Optional[StateDescriptor]
     output: Optional[Type[Features]]
+    function: Callable[P, T]
     function_definition: str
-    function: Callable
     doc: Optional[str]
     default_args: List[Optional[ResolverArgErrorHandler]]
 
 
-class ResolverAnnotationParser:
-    def __init__(self, resolver: Callable, glbs: Optional[Dict[str, Any]], lcls: Optional[Dict[str, Any]]):
-        self.resolver = resolver
-        self.glbs = glbs
-        self.lcls = lcls
-
-        self._args = {arg.arg: arg for arg in self._get_resolver_args()}
-
-    def _get_resolver_args(self):
-        source = inspect.getsource(self.resolver)
-        parsed_source = ast.parse(textwrap.dedent(source))
-        assert len(parsed_source.body) == 1
-        function_def = parsed_source.body[0]
-        assert isinstance(function_def, ast.FunctionDef)
-        args = function_def.args
-        all_args = [*args.posonlyargs, *args.args, *args.kwonlyargs]
-        return all_args
-
-    def parse_annotation(self, name: str):
-        arg = self._args[name]
-        annotation = arg.annotation
-        assert annotation is not None
-        if isinstance(annotation, ast.Constant):
-            val = annotation.value
-            assert isinstance(val, str), "if a literal annotation, it must be a string"
-            # string of type annotation
-            val = ast.parse(val, mode="eval")
-            assert isinstance(val, ast.Module)
-            annotation = val.body
-        if isinstance(annotation, ast.Subscript):
-            # All fancy ast parsing would appear within the subscript of a df __getitem__
-            annotation = ast.Subscript(
-                value=annotation.value,
-                slice=convert_slice(annotation.slice),
-                ctx=annotation.ctx,
-            )
-        return eval_converted_expr(annotation, self.glbs, self.lcls)
-
-
 def get_resolver_fqn(function: Callable):
     return f"{function.__module__}.{function.__name__}"
 
@@ -343,8 +308,10 @@ def get_state_default_value(
 ) -> Any:
     if not issubclass(state_typ, BaseModel) and not dataclasses.is_dataclass(state_typ):
         raise ValueError(
-            f"State value must be a pydantic model or dataclass, "
-            f"but argument '{parameter_name_for_errors}' has type '{type(state_typ).__name__}'"
+            (
+                f"State value must be a pydantic model or dataclass, "
+                f"but argument '{parameter_name_for_errors}' has type '{type(state_typ).__name__}'"
+            )
         )
 
     default = declared_default
@@ -372,17 +339,23 @@ def get_state_default_value(
 
 
 def parse_function(
-    fn: Callable,
+    fn: Callable[P, T],
     glbs: Optional[Dict[str, Any]],
     lcls: Optional[Dict[str, Any]],
     ignore_return: bool = False,
-) -> ResolverParseResult:
+    allow_custom_args: bool = False,
+    is_streaming_resolver: bool = False,
+) -> ResolverParseResult[P, T]:
     fqn = get_resolver_fqn(function=fn)
     sig = inspect.signature(fn)
     annotation_parser = ResolverAnnotationParser(fn, glbs, lcls)
 
     function_definition = inspect.getsource(fn)
-    return_annotation = get_type_hints(fn)["return"]
+    return_annotation = get_type_hints(fn).get("return")
+    if return_annotation is None:
+        raise TypeError(
+            f"Resolver '{fqn}' must have a return annotation. See https://docs.chalk.ai/docs/resolver-outputs for more information."
+        )
 
     ret_val = None
 
@@ -447,19 +420,29 @@ def parse_function(
             default_args[i] = ResolverArgErrorHandler(parameter.default)
 
         if not isinstance(arg, (StateWrapper, Feature)):
+            if allow_custom_args:
+                continue
             raise bad_input()
 
+        if isinstance(arg.typ, ParsedAnnotation) and isinstance(arg.typ.underlying, Windowed):
+            f = arg.typ.underlying
+            raise ValueError(
+                (
+                    f"Resolver argument '{arg_name}' to '{fqn}' does not select a window. "
+                    f"Add a selected window, like {arg.name}('{next(iter(f._buckets), '')}'). "
+                    f"Available windows: {', '.join(f._buckets)}."
+                )
+            )
+
         if not isinstance(arg, StateWrapper):
             continue
 
-        typ = arg.typ
         if state is not None:
             raise ValueError(
                 f"Only one state argument is allowed. Two provided to '{fqn}': '{state.kwarg}' and '{arg_name}'"
             )
 
         arg_name = parameter.name
-        arg = inputs[i]
 
         state = StateDescriptor(
             kwarg=arg_name,
@@ -470,18 +453,36 @@ def parse_function(
                 parameter_name_for_errors=arg_name,
                 declared_default=parameter.default,
             ),
-            typ=typ,
+            typ=arg.typ,
         )
 
     assert ret_val is None or issubclass(ret_val, Features)
-
+    if not ignore_return and ret_val is not None and issubclass(ret_val, Features):
+        # Streaming resolvers are themselves windowed, so the outputs must not specify a window explicitly.
+        for f in _flatten_features(ret_val):
+            if f.is_windowed and not is_streaming_resolver:
+                raise TypeError(
+                    (
+                        "Windowed features need to specify a window "
+                        f"period in the return type. The feature "
+                        f"'{f.root_fqn}' returned from '{fn.__name__}' does not."
+                    )
+                )
+            if f.is_windowed_pseudofeature and is_streaming_resolver:
+                feature_name_without_duration = "__".join(f.root_fqn.split("__")[:-1])  # A bit hacky, but should work
+                raise TypeError(
+                    (
+                        "Streaming resolvers should not resolve features of a particular window period in the return type"
+                        f"Resolver '{fn.__name__}' returned feature '{f.root_fqn}'. Instead, return {feature_name_without_duration}"
+                    )
+                )
     state_index = state.pos if state is not None else None
     return ResolverParseResult(
         fqn=fqn,
         inputs=[v for i, v in enumerate(inputs) if i != state_index],
         output=ret_val,
-        function_definition=function_definition,
         function=fn,
+        function_definition=function_definition,
         doc=fn.__doc__,
         state=state,
         default_args=default_args,
@@ -523,7 +524,7 @@ def online(
     caller_globals = caller_frame.frame.f_globals
     caller_locals = caller_frame.frame.f_locals
 
-    def decorator(fn: Callable[P, T], cf: str = caller_filename):
+    def decorator(fn: Callable[P, T]):
         parsed = parse_function(fn, caller_globals, caller_locals)
         if not env_var_bool("CHALK_ALLOW_REGISTRY_UPDATES") and parsed.fqn in {s.fqn for s in Resolver.registry}:
             raise ValueError(f"Duplicate resolver {parsed.fqn}")
@@ -531,13 +532,13 @@ def online(
             raise ValueError(f"Online resolvers must return features; '{parsed.fqn}' returns None")
 
         resolver = OnlineResolver(
-            filename=cf,
+            filename=caller_filename,
             function_definition=parsed.function_definition,
             fqn=parsed.fqn,
             doc=parsed.doc,
             inputs=parsed.inputs,
             output=parsed.output,
-            fn=parsed.function,
+            fn=fn,
             environment=None if environment is None else list(ensure_tuple(environment)),
             tags=None if tags is None else list(ensure_tuple(tags)),
             max_staleness=None,
@@ -606,7 +607,7 @@ def offline(
             doc=parsed.doc,
             inputs=parsed.inputs,
             output=parsed.output,
-            fn=parsed.function,
+            fn=fn,
             environment=None if environment is None else list(ensure_tuple(environment)),
             tags=None if tags is None else list(ensure_tuple(tags)),
             max_staleness=None,
@@ -666,6 +667,7 @@ def sink(
 
     def decorator(fn: Callable[P, T], cf: str = caller_filename):
         parsed = parse_function(fn, caller_globals, caller_locals, ignore_return=True)
+        assert parsed.output is not None
         SinkResolver.registry.append(
             SinkResolver(
                 filename=cf,
@@ -674,7 +676,7 @@ def sink(
                 doc=parsed.doc,
                 inputs=parsed.inputs,
                 output=parsed.output,
-                fn=parsed.function,
+                fn=fn,
                 environment=None if environment is None else list(ensure_tuple(environment)),
                 tags=None if tags is None else list(ensure_tuple(tags)),
                 machine_type=machine_type,
@@ -688,3 +690,245 @@ def sink(
         return fn
 
     return decorator(fn) if fn else decorator
+
+
+class StreamResolver(Resolver[P, T]):
+    registry: "List[StreamResolver]" = []
+
+    def __init__(
+        self,
+        caller_globals: Optional[Dict[str, Any]],
+        caller_locals: Optional[Dict[str, Any]],
+        caller_filename: str,
+        source: StreamSource,
+        fn: Callable[P, T],
+        mode: Optional[str],
+        signature: StreamResolverSignature,
+        environment: Optional[List[str]],
+        machine_type: Optional[str],
+        output: Type[Features],
+    ):
+        parsed = parse_function(fn, caller_globals, caller_locals, allow_custom_args=True, is_streaming_resolver=True)
+        self.function_definition = parsed.function_definition
+        self.fqn = parsed.fqn
+        self.filename = caller_filename
+        self.source = source
+        self.mode = mode
+        self.fn = fn
+        self.environment = environment
+        self.doc = parsed.doc
+        self.inputs = []
+        self.__doc__ = parsed.doc
+        self.__module__ = fn.__module__
+        self.machine_type = machine_type
+        self.max_staleness = None
+        self.state = parsed.state
+        self.tags = None  # TODO -- should there be tags?
+        self.signature = signature
+        self.output = output
+        self.default_args = []
+        fqn_to_windows = {o.fqn: o.window_durations for o in _flatten_features(self.output) if o.is_windowed}
+        if len(set(tuple(v) for v in fqn_to_windows.values())) > 1:
+            fqn_to_declared_windows = {
+                o.fqn: sorted(o.window_durations) for o in _flatten_features(self.output) if o.is_windowed
+            }
+            periods = [f'{fqn}[{", ".join(f"{window}s")}]' for fqn, window in fqn_to_declared_windows.items()]
+            raise ValueError((f"All features must have the same window periods. Found " f"{', '.join(periods)}"))
+        self.window_periods_seconds = next(iter(fqn_to_windows.values()), ())
+        # Mapping of window (in secs) to mapping of (original feature, windowed pseudofeature)
+        self.windowed_pseudofeatures: Dict[int, Dict[Feature, Feature]] = {}
+        for window_period in self.window_periods_seconds:
+            self.windowed_pseudofeatures[window_period] = {}
+            for o in _flatten_features(self.output):
+                if o.is_windowed:
+                    windowed_fqn = get_name_with_duration(o.root_fqn, window_period)
+                    windowed_feature = Feature.from_root_fqn(windowed_fqn)
+                    self.windowed_pseudofeatures[window_period][o] = windowed_feature
+
+    @property
+    def output_features(self):
+        return _flatten_features(self.output)
+
+    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
+        return self.fn(*args, **kwargs)
+
+    def __repr__(self):
+        return f"StreamResolver(name={self.fqn})"
+
+
+def _is_stream_resolver_body_type(annotation: Type):
+    origin = get_origin(annotation)
+    if origin is not None:
+        return False
+    return (
+        isinstance(annotation, type) and issubclass(annotation, (str, bytes, BaseModel))
+    ) or dataclasses.is_dataclass(annotation)
+
+
+def _parse_stream_resolver_param(
+    param: Parameter,
+    annotation_parser: ResolverAnnotationParser,
+    resolver_fqn_for_errors: str,
+    is_windowed_resolver: bool,
+) -> StreamResolverParam:
+    if param.kind not in {Parameter.KEYWORD_ONLY, Parameter.POSITIONAL_OR_KEYWORD}:
+        raise ValueError(
+            f"Stream resolver {resolver_fqn_for_errors} includes unsupported keyword or variadic arg '{param.name}'"
+        )
+
+    annotation = annotation_parser.parse_annotation(param.name)
+    if isinstance(annotation, StateWrapper):
+        if is_windowed_resolver:
+            raise ValueError("Windowed resolvers cannot have state")
+        default_value = get_state_default_value(
+            state_typ=annotation.typ,
+            declared_default=param.default,
+            resolver_fqn_for_errors=resolver_fqn_for_errors,
+            parameter_name_for_errors=param.name,
+        )
+        return StreamResolverParamKeyedState(
+            name=param.name,
+            typ=annotation.typ,
+            default_value=default_value,
+        )
+
+    if not is_windowed_resolver and _is_stream_resolver_body_type(annotation):
+        return StreamResolverParamMessage(name=param.name, typ=annotation)
+
+    if is_windowed_resolver and get_origin(annotation) in (list, List):
+        item_typ = get_args(annotation)[0]
+        if _is_stream_resolver_body_type(item_typ):
+            return StreamResolverParamMessageWindow(name=param.name, item_typ=item_typ)
+
+    raise ValueError(
+        (
+            f"Stream resolver parameter '{param.name}' of resolver '{resolver_fqn_for_errors}' is not recognized. "
+            f"Message payloads must be one of `str`, `bytes`, or pydantic model class. "
+            f"Keyed state parameters must be chalk.KeyedState[T]. "
+            f"Received: {annotation}"
+        )
+    )
+
+
+def parse_stream_resolver_params(
+    user_func: Callable,
+    *,
+    resolver_fqn_for_errors: str,
+    annotation_parser: ResolverAnnotationParser,
+    is_windowed_resolver: bool,
+) -> Sequence[StreamResolverParam]:
+    sig = inspect.signature(user_func)
+    params = [
+        _parse_stream_resolver_param(p, annotation_parser, resolver_fqn_for_errors, is_windowed_resolver)
+        for p in sig.parameters.values()
+    ]
+
+    uses_message_body = any(
+        p for p in params if isinstance(p, (StreamResolverParamMessage, StreamResolverParamMessageWindow))
+    )
+
+    if not uses_message_body:
+        raise ValueError(
+            f"Stream resolver '{resolver_fqn_for_errors}' must take as input "
+            + f"a pydantic model, `str`, or `bytes` representing the message body"
+        )
+
+    keyed_state_params = [p.name for p in params if isinstance(p, StreamResolverParamKeyedState)]
+    if len(keyed_state_params) > 1:
+        raise ValueError(
+            f"Stream resolver '{resolver_fqn_for_errors}' includes more than one KeyedState parameter: {keyed_state_params}"
+        )
+
+    return params
+
+
+def parse_stream_resolver_output_features(
+    user_func: Callable,
+    *,
+    resolver_fqn_for_errors: str,
+) -> Type[Features]:
+    return_annotation = get_type_hints(user_func).get("return")
+    if return_annotation is None:
+        raise TypeError(
+            (
+                f"Resolver '{user_func.__name__}' must have a return annotation. See "
+                f"https://docs.chalk.ai/docs/resolver-outputs for "
+                f"more information."
+            )
+        )
+
+    if not isinstance(return_annotation, type):
+        raise TypeError(f"return_annotation {return_annotation} of type {type(return_annotation)} is not a type")
+
+    if issubclass(return_annotation, DataFrame):
+        return Features[return_annotation]
+
+    if not issubclass(return_annotation, Features):
+        raise ValueError(
+            f"Stream resolver '{resolver_fqn_for_errors}' did not have a valid return type: "
+            + f"must be a features class or chalk.features.Features[]"
+        )
+
+    # TODO: validate that these features are in the same namespace and include a pkey
+    output_features = cast(Type[Features], return_annotation)
+
+    return output_features
+
+
+def parse_and_register_stream_resolver(
+    *,
+    caller_globals: Optional[Dict[str, Any]],
+    caller_locals: Optional[Dict[str, Any]],
+    fn: Callable[P, T],
+    source: StreamSource,
+    caller_filename: str,
+    mode: Optional[str] = None,
+    environment: Optional[Union[List[str], str]] = None,
+    machine_type: Optional[MachineType] = None,
+) -> StreamResolver[P, T]:
+    fqn = f"{fn.__module__}.{fn.__name__}"
+    annotation_parser = ResolverAnnotationParser(fn, caller_globals, caller_locals)
+    output_features = parse_stream_resolver_output_features(
+        fn,
+        resolver_fqn_for_errors=fqn,
+    )
+    flattened_output_features = (
+        df.columns
+        if len(output_features.features) == 1
+        and isinstance(output_features.features[0], type)
+        and issubclass(df := output_features.features[0], DataFrame)
+        else output_features.features
+    )
+    is_windowed_resolver = any(x.is_windowed for x in flattened_output_features)
+    params = parse_stream_resolver_params(
+        fn,
+        resolver_fqn_for_errors=fqn,
+        annotation_parser=annotation_parser,
+        is_windowed_resolver=is_windowed_resolver,
+    )
+
+    if isinstance(output_features.features[0], type) and issubclass(output_features.features[0], DataFrame):
+        output_feature_fqns = set(f.fqn for f in cast(Type[DataFrame], output_features.features[0]).columns)
+    else:
+        output_feature_fqns = set(f.fqn for f in output_features.features)
+
+    signature = StreamResolverSignature(
+        params=params,
+        output_feature_fqns=output_feature_fqns,
+    )
+
+    resolver = StreamResolver(
+        caller_globals=caller_globals,
+        caller_locals=caller_locals,
+        caller_filename=caller_filename,
+        source=source,
+        mode=mode,
+        fn=fn,
+        environment=None if environment is None else list(ensure_tuple(environment)),
+        machine_type=machine_type,
+        signature=signature,
+        output=output_features,
+    )
+
+    StreamResolver.registry.append(resolver)
+    return resolver
