@@ -14,3 +14,7 @@ def removeprefix(text: str, prefix: str):
     but we're supporting down to 3.8, where it is not available.
     """
     return text[len(prefix) :] if text.startswith(prefix) else text
+
+
+def comma_whitespace_split(value: str):
+    return re.split(r"\s*,\s*", value)
