@@ -16,7 +16,7 @@ class KafkaMessage(BaseModel):
     val_a: str
 
 
-source = KafkaSource(bootstrap_server=[], topic=[])
+source = KafkaSource(bootstrap_server="bootstrap-server", topic="topic")
 
 
 @dataclasses.dataclass
